/*-------------------------------------------------------------------------------
Parallel Algorithm for Maximum Flow in Flow Network using OpenMP
Flow Graph represented as Matrix - G[i][j] => flow between ith and jth vertex in graph G  
Author: V.Sai Venkata Krishnan
Input format: ./ProgName
Assuming vertices are numbered 0 to V-1
---------------------------------------------------------------------------------*/

#include<iostream>
#include<vector>
#include<algorithm>
#include<omp.h>
#include<cmath>
#include<fstream>
#include<ctime>
#include<climits>
#include<string>

using namespace std;

vector< vector<int> > G;
vector<long> Distance;

//Pennant - Tree with 2^k vertices k = 0,1,2,....
class Pennant
{	public:
		Pennant *left;
		Pennant *right;
		long data;		
		long size;		

		Pennant()
		{	left = NULL;
			right = NULL;
			size=0;
			data = -1;
		}
			
};

//Bag - array of Pennnants
class Bag
{	public:
		vector<Pennant*> S;
		long size;
		Bag()
		{	S.resize(64); 
			size = 0;		
		}
		
		void operator = (const Bag &B2)
		{	size = B2.size;
			S = B2.S;	
		}
				
};

//Deallocation of Dynamically Allocated Memory
void DeletePennant(Pennant* &x)
{	if(x == NULL)
		return;
	else
	{	DeletePennant(x->left);
		DeletePennant(x->right);
		delete x;
		x = NULL;
		return;
	}	
}	

//Removing Elements/Pennants from Bag
void RemoveElements(Bag &B)
{	
	int k;
	for(k=0 ; k<B.S.size() ; k++)
	{	if(B.S[k]!=NULL)
		{	DeletePennant(B.S[k]);
		}
	}
	B.size = 0;
}

//Union of two Pennants
void PennantUnion(Pennant* &x,Pennant* &y)
{	y->right = x->left;
	x->left = y;
	x->size = x->size + y->size;
}

//Splitting two Pennants
void PennantSplit(Pennant* &x,Pennant* &y)
{	y = x->left;
	x->left = y->right;
	y->right = NULL;
	x->size = x->size / 2;
	y->size = x->size;
}


//Displaying Pennant
void TraversePennant(Pennant* x)
{	if(x == NULL)
		return;
	else
	{	cout<<"\nData = "<<x->data<<"\n";	
		TraversePennant(x->left);
		TraversePennant(x->right);
	}	
}

//Insert element into the Bag i.e. in any one of the Pennants within the Bag 
void BagInsert(Bag &B,long element)
{	int k=0;
	Pennant* x = new Pennant();
	x->data = element;
	x->size = 1;
	while(B.S[k] != NULL)
	{	PennantUnion(x,B.S[k]);
		B.S[k++] = NULL;
	}
	B.S[k] = x;
	B.size++;	
}

//Full Adder implementation for joining Bag Union
void FA(Pennant* &x ,Pennant* y ,Pennant* &z ,int k)
{	if(x == NULL && y != NULL  && z == NULL)
		x = y;
	else if(x == NULL && y == NULL  && z != NULL)
	{	x = z;
		z = NULL;
	}
	else if(y != NULL && z != NULL)
	{	PennantUnion(y,z);
		z = y;
		
	}
	else if(x != NULL && y != NULL && z == NULL)
	{	PennantUnion(x,y);
		z = x;
		x = NULL;
	}
	else if(x != NULL && y == NULL && z != NULL)
	{	PennantUnion(x,z);
		z = x;
		x = NULL;
	}
}

//Union of two Bags
void BagUnion(Bag &B1 ,Bag &B2)
{	Pennant *y = NULL ; //carry bit;
	long k;
	for(k=0 ; k<B1.S.size() ; k++) 
	{	FA(B1.S[k],B2.S[k],y,k);
	}
	B1.size += B2.size; 
}


//Resetting the Distance vector
void NullifyDistance()
{	for(long i=0;i<Distance.size();i++)
		Distance[i] = -1; 
}

//Display Contents of Bag
void DisplayBag(Bag &B)
{	int k;
	for(k=0 ; k<B.S.size() ; k++)
	{	if(B.S[k]!=NULL)
			TraversePennant(B.S[k]);
	}
}


//Testing function  - to test Union and Insert operation w.r.t Bag
void PennantBagTester()
{	Bag B1; //create an empty bag of a constant size
	Bag B2;
	int n,m;
	long data;
	cout<<"\nEnter the no of elements in bag 1\n";
	cin>>n;
	for(int i=0;i<n;i++)
	{	cout<<"\nEnter data : ";
		cin>>data;
		BagInsert(B1,data);
	}
	cout<<"\nEnter the no of elements in bag 2\n";
	cin>>m;
	for(int i=0;i<m;i++)
	{	cout<<"\nEnter data : ";
		cin>>data;	
		BagInsert(B2,data);		
	}
	cout<<"\n***********BAG 1************\n";
	DisplayBag(B1);
	cout<<"\nSize = "<<B1.size;
	cout<<"\n***********BAG 2************\n";
	DisplayBag(B2);
	cout<<"\nSize = "<<B2.size;

	cout<<"\n***********BAG 1 UNION BAG 2************\n";
	BagUnion(B1,B2);	
	DisplayBag(B1);	
	cout<<"\nSize = "<<B1.size;

	RemoveElements(B1);
	RemoveElements(B2);

	cout<<"\nAfter deletion\n";
	cout<<"\n***********BAG 1************\n";
	DisplayBag(B1);
	cout<<"\nSize = "<<B1.size;
	cout<<"\n***********BAG 2************\n";
	DisplayBag(B2);
	cout<<"\nSize = "<<B2.size;
	cout<<endl;
}

//Read Graph from file 
void ReadGraph()
{	string graphName;
	cout<<"\nEnter the name of graph with extension\n";
	cin>>graphName;
	ifstream myfile1(graphName,ios_base::in);
	
	long s;
	long d;
	int flow;
	long n;
	int m;
	long i;
	long j;

	cout<<"\nEnter the no of nodes\n";
	cin>>n;
	
	G.resize(n); 
	
	for(i=0;i<n;i++)
	{	G[i].resize(n);
		
		for(j=0;j<n;j++)
		{	G[i][j]=0;
				
		}
	}
	while(myfile1>>s)
	{	myfile1>>d;
		myfile1>>flow;
		//G[s].push_back(make_tuple(v,flow));		
		G[s][d]=flow;
	}	
			
}

//To display the distance vertex(level of vertex from the source) which is computed using BFS 
void DisplayDistance()
{	for(int i=0;i<Distance.size();i++)
		cout<<"\nVertex "<<i<<" : "<<Distance[i];
}

//To store all the data elements of the Pennant in a vector
void GetListPennant(Pennant *inPennant , vector<long> &inBagData)
{	if(inPennant == NULL)
		return;
	else
	{	inBagData.push_back(inPennant->data);	
		GetListPennant(inPennant->left , inBagData);
		GetListPennant(inPennant->right , inBagData);
	}	
}

//Process the Pennant data elements
void ProcessPennant(Pennant *inPennant , Bag &outBag , long d)
{	long u;	
	if(inPennant->size >= 0) //grainsize  = 4
	{	vector<long> inBagData;
		GetListPennant(inPennant , inBagData);
		for(int i=0 ; i<inBagData.size() ; i++)
		{	
			#pragma omp parallel for shared(i,inBagData,Distance,d,G,outBag) private(u) num_threads(4)
			for(u=0;u<G.size();u++)
			{	//cout<<endl<<inBagData[i]<<" "<<u<<" "<<G[i][u];	
				if(G[inBagData[i]][u]>0)
				{	if(Distance[u] == -1)
					{	Distance[u] = d+1; //benign data race - So no need of Critical construct
						BagInsert(outBag,u);				
					}
				}
			}
		}
		return;
	}
	/*
	Pennant *newPennant = NULL;
	PennantSplit(inPennant,newPennant);
	Bag tempOutBag;
	
	#pragma omp parallel 
	{
		#pragma omp sections nowait private(tempOutBag) 
		{	
			#pragma omp section 		
			{	ProcessPennant(inPennant,tempOutBag,d);
			}
			#pragma omp section
			{	ProcessPennant(inPennant,tempOutBag,d);
			}
		}
		#pragma omp critical
		{	BagUnion(outBag,tempOutBag);
		}
		#pragma omp barrier
	}
	*/ 
	
}

//Process Each layer of the graph
void ProcessLayer(Bag inBag , Bag &outBag , long d  )
{	int k;	
	Bag tempOutBag;
	int size = floor(log2(inBag.size));
	#pragma omp parallel shared(inBag , size , outBag , d) private(tempOutBag) num_threads(4)
	{	
		#pragma omp for  nowait
		for(k=0 ; k <= size ; k++)
		{	if(inBag.S[k] != NULL)
			{	ProcessPennant(inBag.S[k] , tempOutBag , d);
			}
		}
		#pragma omp critical
		{	BagUnion(outBag,tempOutBag);
		}
		#pragma omp barrier
	}		
}

//Parallel BFS - Level Order traversal of the graph using Bags and Pennants
long ParallelBFS(long s,long dest)
{	long d;
	long i;
	#pragma omp parallel for shared(Distance) private(i) num_threads(4)
	for(i=0;i<G.size();i++)
		Distance[i] = -1;
	Distance[s] = 0; // setting source level to zero 
	d = 0;	
	Bag V0;
	BagInsert(V0,s);
	while(V0.size > 0)
	{	
		Bag V1;
		ProcessLayer(V0,V1,d);
		RemoveElements(V0);
		V0 = V1;
		d++;
	}			
	return (Distance[dest]);
}

//To find all possible paths between the source and destination
void FindPath(long s,long dest,long level ,vector<long> &path,vector< vector<long> > &allPaths)
{	path.push_back(dest);
	level --;
	if(level>0)	
	{	for(long i=0;i<Distance.size();i++)
		{	if(Distance[i] == level && G[i][dest]>0)
			{	FindPath(s,i,level,path,allPaths);
				path.pop_back();		
			}
		}
	}
	else
	{	path.push_back(s);
		allPaths.push_back(path);	
	}	
			
}

//To display all the paths found between the source and destination
void DisplayAllPaths(vector< vector<long> > &allPaths)
{	for(int i=0;i<allPaths.size();i++)
	{	cout<<"\n";
		for(int j=allPaths[i].size()-1;j>=0;j--)
		{	cout<<allPaths[i][j]<<"->";
		}
	}
}

//To calculate the maximum flow between source and destination
long MaximumFlow(long s , long dest)
{	long maxFlow = 0;	
	while(ParallelBFS(s,dest) > 0)
	{	vector<long> path;
		vector< vector<long> > allPaths;
		
		FindPath(s,dest,Distance[dest],path,allPaths);
		//DisplayAllPaths(allPaths);
		
		long pathFlow;
		int flag = 1;		
		for(int i=0;i<allPaths.size();i++)
		{	pathFlow = LONG_MAX;
			for(int j=0;j < allPaths[i].size()-1 ; j++)
			{	if(G[allPaths[i][j+1]][allPaths[i][j]] > 0)
				{	pathFlow = (pathFlow<=G[allPaths[i][j+1]][allPaths[i][j]]) ? 
							pathFlow:G[allPaths[i][j+1]][allPaths[i][j]];
				
				}
				else
				{	flag = 0;
					break;	
				}	
			}	
			if(flag == 1)
			{	for(int j=0;j < allPaths[i].size()-1 ; j++)
				{	G[allPaths[i][j+1]][allPaths[i][j]] -= pathFlow;
					G[allPaths[i][j]][allPaths[i][j+1]] += pathFlow;
				}
				maxFlow += pathFlow;
					
			}	
			flag = 1;		
		}
//		cout<<"\nMax flow = "<<maxFlow;
		NullifyDistance();
	}
	
	return maxFlow;
}

//To save the Distance Vector computed using BFS
void SaveDistance()
{	ofstream myfile("DistanceParallel1.txt");
	for(int i=0;i<Distance.size();i++)
	{	myfile<<Distance[i]<<"\n";
	}
	myfile.close();
}

int main()
{
//	PennantBagTester();

	long s;
	long d;
	ReadGraph();
	Distance.resize(G.size(),-1);

	cout<<"\nEnter source vertex id :\n";
	cin>>s;
	cout<<"\nEnter destination vertex id :\n";
	cin>>d;
	clock_t t;
	t = clock();	

//	bool r=ParallelBFS(s,d);
	long maxFlow;
	maxFlow = MaximumFlow(s,d); 
	t = clock() -t ;
	double timeTaken = ((double)t)/CLOCKS_PER_SEC;
	cout<<"\nMaximum flow is "<<maxFlow<<endl;
	cout<<"\nTime taken = "<<timeTaken;			
//	DisplayDistance();
//	cout<<"\nr="<<r;
//	SaveDistance();
	return 0;
}


